// By Warren for Website "TopQuantumFusion"
// Cohens Algorithm 1.5.2 Cornacchia 
// Includes
// H.Cohen p.29  Algorithm 1.4.12 Binary Kronecker
// Includes
// 'Sudhir Sharma'
// Tonelli_Shanks_Sqr
// As checked for 'Henry Cohen' example x^2 + 2y^2 = 97
// page 35 "A course in computational algebraic number theory"
// Note Henry Cohen for Complex quadratic orders of discriminant
// D < 0, congruent to 0 or 1 Mod 4. Suggests Algorithm 1.5.3 (Modified Cornacchia)
// Over did it with the "DP4" and could use "Try Catch" Error coding.
// checked again 12/11/2023
#include <iostream>
#include <cmath>
#include <math.h>
using namespace std;
typedef long long int DP4;

DP4 powerMod(DP4 base, DP4 exponent, DP4 modulus) {
   DP4 result = 1;
   base = base % modulus;
   while (exponent > 0) {
      if (exponent % 2 == 1)
      result = (result * base)% modulus;
      exponent = exponent >> 1;
      base = (base * base) % modulus;
   }
   return result;
}
DP4 gcd(DP4 a, DP4 b) {
   if (b == 0)
   return a;
   else
   return gcd(b, a % b);
}
DP4 orderValues(DP4 p, DP4 b) {
   if (gcd(p, b) != 1) {
      return -1;
   }
   DP4 k = 3;
   while (1) {
      if (powerMod(b, k, p) == 1)
      return k;
      k++;
   }
}
DP4 findx2e(DP4 x, DP4& e) {
   e = 0;
   while (x % 2 == 0) {
      x /= 2;
      e++;
   }
   return x;
}
DP4 Tonelli_Shanks_Sqrt(DP4 n, DP4 p) {
   if (gcd(n, p) != 1) {
      return -1;
   }
   if (powerMod(n, (p - 1) / 2, p) == (p - 1)) {
      return -1;
   }
   DP4 s, e;
   s = findx2e(p - 1, e);
   DP4 q;
   for (q = 2; ; q++) {
      if (powerMod(q, (p - 1) / 2, p) == (p - 1))
      break;
   }
   DP4 x = powerMod(n, (s + 1) / 2, p);
   DP4 b = powerMod(n, s, p);
   DP4 g = powerMod(q, s, p);
   DP4 r = e;
   while (1) {
      DP4 m;
      for (m = 0; m < r; m++) {
         if (orderValues(p, b) == -1)
         return -1;
         if (orderValues(p, b) == pow(2, m))
         break;
      }
      if (m == 0)
      return x;
      x = (x * powerMod(g, pow(2, r - m - 1), p)) % p;
      g = powerMod(g, pow(2, r - m), p);
      b = (b * g) % p;
      if (b == 1)
      return x;
      r = m;
   }
}

DP4 Binary_Kronecker(long long a, long long b)
{
	int Pal_2s[]{ 0, 1, 0, -1, 0, -1, 0, 1 };// (-1)^((a^2-1)/8) == tab2[a & 7]
	DP4 v = 0;
	DP4 k = 0;
	DP4 r = 0;
	bool correct2{}; // Initialised to zero
	bool proper1{}; // so is false
	//Step 1
	do{
		if ((b % 2) == 0) {
			cout << " Even....Input 'b' is an odd prime ? ";
			return b;
			break; }// proper1 = true; }odd prime? still true if p and q positive odd integer
	    if ((b == 0) && (abs(a) != 1)) { return 0; }
	    if ((b == 0) && (abs(a) == 1)) { return 0; }

		if (((a % 2) == 0) && ((b % 2) == 0)) { return 0; proper1 = true;} //Terminate Program
		else if ((b % 2) == 0) {
			v += 1;
			b /= 2; 
		}
		if ((v % 2) == 0) {
			k = 1;
		}
		else
		{
			k *= Pal_2s[a&7]; // *=  
		}
		if (b < 0) {
			b = -b;
		}
		if ((b < 0) && (a < 0)) {
			k = -k;
		}
		//Step 3
		if ((b & 1) && (b > 0)) {
			a = a % b;
		}
		//Step 4
		do {
			if ((a == 0) && (b > 1)) {
				return 0;
				proper1 = true;
				correct2 = true;
			}
			if ((a == 0) && (b == 1)) {
				return k;
				proper1 = true;
				correct2 = true;
			}   //Terminate Program
		//Step 5
			v = 0;
			if ((a % 2) == 0) {
				v += 1;
				a = (a / 2);
				if (v & 1) {
					k *= Pal_2s[b & 7]; 
				}
			}
				//Step 6
			if (((b % 2) == 1) && (a % 2) == 1){   ///((b & 1) && (a & 1)) {
				r = b - a;
				if (r > 0) {
					if ((a&b&2) != 0) { k = -k; }
					b = a;
					a = r;
				}
				else {
					a = -r;
				} //Go to Step 4
			}
			
		} while (correct2 == false);
	} while (proper1 == false);
}
int main() {
	DP4 an = 0;
	DP4 bp = 0;
	DP4 kron = 0;
	DP4 Xon = 0;
	DP4 Eon = 0;
	DP4 Eob = 0;
	DP4 lob = 0;
	DP4 Rob = 0;
	DP4 Ceat = 0;
	DP4 SqrtNum1 = 0;
	DP4 XNum1 = 0;
	DP4 YNum1 = 0;
	cout << "Enter 'integer' less than the chosen prime" << endl;
	cin >> an;
	cout << "Enter number typically a prime" << endl;
	cin >> bp;
	kron = Binary_Kronecker(an, bp);
	if (kron == -1){
		cout << "Not Valid, Quadratic non residual" << endl;
		}
		else{
			Xon = Tonelli_Shanks_Sqrt(an, bp);
		}
	if (Xon == -1){
      cout<<"Modular square root does not exist";}
	  else{
		cout<<"Modular square root of the number is "<< Xon << endl;
	  }
	Xon = abs(Xon - (kron * bp));
	//cout<<"Xon"<< Xon <<endl;
	if(bp/2 < Xon < bp){
	Eon = bp; //prime
	Eob = Xon;
	lob = sqrt(bp);}
	else{cout<< "Adjusted Modular Square Root, not in prime range"<< endl;}
	while (Eob > lob) {
		Rob = Eon % Eob;
		Eon = Eob;
		Eob = Rob;}
	Ceat = ((bp - (Eob * Eob))/ an);
	SqrtNum1 = sqrt(Ceat);
	if(SqrtNum1 == floor(SqrtNum1)){
		XNum1 = Eob;
		YNum1 = SqrtNum1;
		cout<< " x  = " << XNum1 << endl;
		cout<< " y  = " << YNum1 << endl;}
		else{cout<<"No unique solution, not a square";} 
}
